from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import compas
import compas.plugins


@compas.plugins.plugin(category='install')
def installable_rhino_packages():
    return ['compas_cem']


# @compas.plugins.plugin(category='install')
# def after_rhino_install(installed_packages):
#     if 'compas_cem' not in installed_packages:
#         return []

#     import compas_fab.rhino.components.install
#     return compas_fab.rhino.components.install.install()


# @compas.plugins.plugin(category='install')
# def after_rhino_uninstall(installed_packages):
#     if 'compas_cem' not in installed_packages:
#         return []

#     import compas_fab.rhino.components.install
#     return compas_fab.rhino.components.install.uninstall()


if __name__ == "__main__":
    pass



import compas

if compas.RHINO:
    from compas_rhino.geometry import RhinoLine


__all__ = ["Edge"]

# ==============================================================================
# Edge
# ==============================================================================


class Edge(object):
    """
    The edge base class.
    """
    def __init__(self, u, v):
        self.u = u
        self.v = v
        self.attributes = {}

    @classmethod
    def from_line(cls, line, *args, **kwargs):
        """
        Create an edge from a line described by two xyz coordinates.

        Parameters
        ----------
        line : ``tuple`` or ``list``
            The xyz coordinates of the line's two end points.
        *args : ``tuple``
            Additional arguments.
        **kwargs : ``dict``
            Extra keyword arguments.

        Returns
        -------
        edge : ``Edge``
            An edge object.
        """
        edge = cls(line[0], line[1], *args, **kwargs)
        return edge

    @classmethod
    def from_rhino_line(cls, rhino_line, *args, **kwargs):
        """
        Create an edge from a rhino line.

        Parameters
        ----------
        rhino_line : ``Rhino.Geometry.Line``
            A rhino line.
        *args : ``tuple``
            Additional arguments.
        **kwargs : ``dict``
            Extra keyword arguments.

        Returns
        -------
        edge : ``Edge``
            An edge object.
        """
        line = RhinoLine.from_geometry(rhino_line).to_compas()
        return cls.from_line(line, *args, **kwargs)

    def __iter__(self):
        """
        Iterates over the start and end nodes of an edge.

        Yields
        ------
        key : ``int``
            The next node key.
        """
        for node in (self.u, self.v):
            yield node


if __name__ == "__main__":
    pass



from compas.geometry import distance_point_point

from compas_cem.optimization.constraints import Constraint


__all__ = ["DeviationEdgeLengthConstraint"]


class DeviationEdgeLengthConstraint(Constraint):
    """
    Make a deviation edge reach a target length.
    """
    def __init__(self, edge=None, length=None, weight=1.0):
        # TODO: needs different serialization mechanism
        super(DeviationEdgeLengthConstraint, self).__init__(edge, length, weight)

    def error(self, data):
        """
        The length error of a deviation edge.

        Returns
        -------
        error : ``float``
            The squared difference bewteen the current and the target length.
        """
        length_a = self.reference(data)
        length_b = self.target()

        diff = length_a - length_b

        return diff * diff * self.weight

    def reference(self, data):
        """
        """
        u, v = self.key()
        point_a = data["node_xyz"][u]
        point_b = data["node_xyz"][v]

        return distance_point_point(point_a, point_b)


if __name__ == "__main__":
    pass



"""
compas_cem.optimization
****************************

.. currentmodule:: compas_cem.optimization.constraints


Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:


Functions
=========

.. autosummary::
    :toctree: generated/
    :nosignatures:

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


# from .<module> import *
from .constraint import *  # noqa F403
from .point import *  # noqa F403
from .plane import *  # noqa F403
from .line import *  # noqa F403
from .force import *  # noqa F403
from .length import *  # noqa F403

# import compas
# if not compas.IPY:
#     from .mesh import *

__all__ = [name for name in dir() if not name.startswith('_')]



from time import time


__all__ = ["solve_nlopt_proxy"]


# ------------------------------------------------------------------------------
# Optimization
# ------------------------------------------------------------------------------

def solve_nlopt_proxy(topology, constraints, parameters, algorithm, iters, eps):
    """
    Deprecated version of ``optimizer_solve_nlopt_proxy``.
    """
    from compas_cem.optimization import Optimizer

    optimizer = Optimizer()

    # add parameters
    for parameter in parameters:
        optimizer.add_goal(parameter)

    # add constraints
    for constraint in constraints:
        optimizer.add_constraint(constraint)

    start = time()
    form = optimizer.solve_nlopt(topology, algorithm, iters, eps)

    duration = time() - start
    x_opt = optimizer.parameters
    l_opt = optimizer.penalty
    evals = optimizer.evals

    return form, x_opt, l_opt, evals, duration


# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    pass



from compas.geometry import closest_point_on_plane
from compas.geometry import distance_point_point_sqrd

from compas_cem.optimization.constraints import Constraint


__all__ = ["PlaneConstraint"]


class PlaneConstraint(Constraint):
    """
    Pulls the xyz position of a node to a target plane.
    """
    def __init__(self, node=None, plane=None, weight=1.0):
        super(PlaneConstraint, self).__init__(node, plane, weight)

    def error(self, data):
        """
        The error between the xyz coords of a node and its closest point on a plane.

        Returns
        -------
        error : ``float``
            The squared distance between the two points.
        """
        point_a = self.reference(data)
        point_b = self.target(point_a)

        return distance_point_point_sqrd(point_a, point_b) * self.weight

    def reference(self, data):
        """
        """
        a = data["node_xyz"][self.key()]

        return a

    def target(self, point):
        """
        """
        plane = self._target

        return closest_point_on_plane(point, plane)


if __name__ == "__main__":
    pass



import matplotlib.pyplot as plt

from math import copysign
from math import fabs

from compas.geometry import add_vectors
from compas.geometry import length_vector
from compas.geometry import scale_vector
from compas.geometry import normalize_vector
from compas.geometry import translate_points

from compas.utilities import geometric_key
from compas_plotters import NetworkPlotter

from compas_cem import COLORS


__all__ = ["FormPlotter"]


class FormPlotter(NetworkPlotter):
    """
    A plotter tailored to draw form-related matters.

    Parameters
    ----------
    form_diagram : :class:`compas_cem.diagrams.FormDiagram`
        The form diagram to plot.
    """
    def __init__(self, form_diagram, *args, **kwargs):
        super(FormPlotter, self).__init__(form_diagram, *args, **kwargs)

        self._edge_colors = {"trail": COLORS["edge"],
                             "deviation": COLORS["edge"]}

        self._edge_state_colors = {-1: COLORS["compression"],
                                   1: COLORS["tension"],
                                   0: COLORS["edge"]}

        self._form = self.datastructure
        self._float_precision = "3f"

    @property
    def form(self):
        """
        The form diagram to plot.

        Returns
        -------
        form : :class:`compas_cem.diagrams.FormDiagram`
            A form diagram.
        """
        return self._form

    @property
    def edge_state_colors(self):
        """
        The default colors to draw edges based on the forces acting on them.

        Returns
        -------
        edge_state_colors : ``dict`` of ``color``
            For compression forces, blue (0, 0, 255).
            For tension force, red (255, 0, 0).
            No force, black (0, 0, 0).
        """
        return self._edge_state_colors

    @property
    def float_precision(self):
        """
        The default decimal precision to render float values as text.

        Returns
        -------
        precision: ``str``
            The float precision value. Defaults to ``3f``.
        """
        return self._float_precision

    def save(self, filepath, tight=True, autoscale=True, bbox_inches="tight", pad_inches=0.0, **kwargs):
        """
        Saves the plot to a file.

        Parameters
        ----------
        filepath : str
            Full path of the file.
        """
        if autoscale:
            self.axes.autoscale(tight=tight)

        if tight:
            plt.tight_layout()
            kwargs_tight = {"bbox_inches": bbox_inches, "pad_inches": pad_inches}
            kwargs.update(kwargs_tight)

        plt.savefig(filepath, **kwargs)

    def draw_nodes(self, *args, **kwargs):
        """
        Draws the nodes of a ``FormDiagram``.

        Parameters
        ----------
        keys : ``list`` of ``int``
            The keys of the nodes to plot.
        radius : ``float``, ``dict`` of ``float``
            The radius of the nodes.
        text : ``str``, ``dict`` of ``str``
            A dictionary of strings to render on the nodes.
        facecolor : ``color``, ``dict`` of ``color``
            Color for the node circle fill in (r, g, b) format.
        edgecolor : ``color``, ``dict`` of ``color``
            Color for the node circle edge in (r, g, b) format.
        edgewidth : ``float``, ``dict`` of ``float``
            Width for the node circle edge.
        textcolor : ``color``, ``dict`` of ``color``
            Color for the text to be displayed on the nodes.
        fontsize : ``int``, ``dict`` of ``int``
            Font size for the text to be displayed on the nodes.

        Returns
        -------
        collection : ``matplotlib.collection``
            A matplotlib point collection object.

        Notes
        -----
        When the parameters are passed as single value, this will be applied to
        all the nodes or edges in the ``FormDiagram``. If instead, a dictionary
        that maps ``{node_key: attribute}`` is supplied, specific values can be
        assigned individually.
        """
        fc = COLORS["node"]

        text = kwargs.get("text")
        if text and text != "key":
            kwargs["text"] = self._text_labels_nodes(text)

        super(FormPlotter, self).draw_nodes(facecolor=fc, *args, **kwargs)

    def draw_edges(self, *args, **kwargs):
        """
        Draws the edges of a ``FormDiagram``.

        Parameters
        ----------
        keys : ``list`` of ``tuple``
            The keys of the edges to plot.
        width : ``float``, ``dict`` of ``float``
            Width of the edges.
        color : ``color``, ``dict`` of ``color``
            Color of the edges in (r, g, b) format.
        text : ``str``, ``dict`` of ``str``
            A dictionary of strings to render on the nodes.
        textcolor : ``color``, ``dict`` of ``color``
            Color for the text to be displayed on the nodes.
        fontsize : ``int``, ``dict`` of ``int``
            Font size for the text to be displayed on the nodes.

        Returns
        -------
        collection : ``matplotlib.collection``
            A matplotlib point collection object.

        Notes
        -----
        When the parameters are passed as single value, this will be applied to
        all the nodes or edges in the ``FormDiagram``. If instead, a dictionary
        that maps ``{edge_key: attribute}`` is supplied, specific values can be
        assigned individually.
        """
        ds = self.datastructure
        cmap = self.edge_state_colors
        ec = {e: cmap[copysign(1.0, ds.edge_attribute(e, "force") or 0.0)] for e in ds.edges()}

        text = kwargs.get("text")
        if text and text != "key":
            kwargs["text"] = self._text_labels_edges(text)

        return super(FormPlotter, self).draw_edges(color=ec, *args, **kwargs)

    def draw_loads(self, keys=None, scale=1.0, width=1.0, gap=0.05, tol=1e-3):
        """
        Draws the node loads in a ``FormDiagram`` as scaled arrows.

        Parameters
        ----------
        keys : ``list``
            The list of node keys where to draw forces.
            If nodes is ``None``, all nodes in will be considered.
            Defaults to ``None``.
        scale : ``float``
            The scale of the load arrows. Defaults to ``1.0``.
        width : ``float``
            The arrows uniform display width. Defaults to ``1.0``.
        gap : ``float``
            The offset between the node and the load. Defaults to ``0.2``.
        tol : ``float``
            The minimum force magnitude to draw. Defaults to ``1e-3``.
        """
        if not keys:
            keys = list(self.datastructure.nodes())
        attrs = ["qx", "qy", "qz"]
        color = COLORS["load"]
        shift = {key: False for key in keys}
        self._draw_forces(keys, attrs, scale, color, width, shift, gap, tol)

    def draw_reactions(self, keys=None, scale=1.0, width=1.0, gap=0.05, tol=1e-3):
        """
        Draws the support reaction forces in a ``FormDiagram`` as scaled arrows.

        Parameters
        ----------
        keys : ``list``
            The list of node keys where to draw forces.
            If nodes is ``None``, all nodes in will be considered.
            Defaults to ``None``.
        scale : ``float``
            The scale of the residual arrows. Defaults to ``1.0``.
        width : ``float``
            The arrows uniform display width. Defaults to ``1.0``.
        gap : ``float``
            The offset between the node and the force. Defaults to ``0.2``.
        tol : ``float``
            The minimum force magnitude to draw. Defaults to ``1e-3``.
        """
        if not keys:
            keys = list(self.datastructure.support_nodes())
        attrs = ["rx", "ry", "rz"]
        color = COLORS["support_force"]

        # TODO: needs a more robust check for arrow orientation
        # what we need is to know whether the arrow needs a full shift.
        # here we say we shift if the connected trail edge is in compression
        form = self.datastructure
        shift = {}
        for key in keys:
            # every support must connect to only one trail edge
            s = False
            forces = [form.edge_force(e) for e in form.connected_edges(key)]
            max_force = max(forces, key=lambda f: fabs(f))
            if max_force < 0.0:
                s = True
            shift[key] = s

        self._draw_forces(keys, attrs, scale, color, width, shift, gap, tol)

    def _draw_forces(self, keys, attrs, scale, color, width, shift, gap, tol):
        """
        Base method to draws forces (residuals or loads) as scaled arrows.

        Parameters
        ----------
        keys : ``list``
            The list of node keys where to draw forces.
        attrs : ``list``
            The attribute names of the force vector to draw.
        scale : ``float``
            The forces scale factor.
        color : ``tuple``
            The forces' color in rgb.
        width : ``float``
            The forces  display width.
        shift : ``bool``
            A flat to shift an arrow one length along its own axis.
        gap : ``float``
            The offset between the node and the force.
        tol : ``float``
            The minimum force magnitude to draw.
        """
        ds = self.datastructure
        arrows = []

        for node in keys:
            q_vec = ds.node_attributes(node, attrs)
            q_vec_scaled = scale_vector(q_vec, scale)
            q_vec_norm = normalize_vector(q_vec)
            q_len = length_vector(q_vec)

            if q_len < tol:
                continue

            arrow = {}
            start = ds.node_xyz(node)
            end = add_vectors(start, q_vec_scaled)

            # shift
            gap_arrow = gap
            if shift[node]:
                gap_arrow = (gap + length_vector(q_vec_scaled)) * -1

            gap_vector = scale_vector(q_vec_norm, gap_arrow)
            start, end = translate_points([start, end], gap_vector)

            # create gap
            arrow["start"] = start
            arrow["end"] = end
            arrow["color"] = color
            arrow["width"] = width

            arrows.append(arrow)

        super(FormPlotter, self).draw_arrows(arrows)

    def draw_segments(self, segments, color=(50, 50, 50), width=0.5, ls="--"):
        """
        Draws additional line segments on a ``FormDiagram``.

        Parameters
        ----------
        segments : ``list``
            The line segments as tuples of xyz coordinates.
        color : ``tuple``
            The lines uniform color in rgb. Defaults to gray, ``(40, 40, 40)``.
        width : ``float``
            The lines' uniform display width. Defaults to ``0.5``.
        """
        lines = []
        for segment in segments:
            line = {}
            start, end = segment

            line["start"] = start
            line["end"] = end
            line["color"] = color
            line["width"] = width

            lines.append(line)

        lines = super(FormPlotter, self).draw_lines(lines)
        lines.set_linestyle(ls)

        return lines

    def _text_labels_nodes(self, text_tag):
        """
        Generates text labels to plot on the nodes of a ``FormDiagram``.

        Input
        -----
        text_tag : `str`
            Tag query. Supported tags are: "xyz" and "key-xyz".

        Returns
        -------
        text_labels : ``list``
            A list of text labels
        """
        def gkey_format(x):
            return geometric_key(ds.node_xyz(x), precision)

        def key_gkey_format(x):
            return "{} / {}".format(x, gkey_format(x))

        ds = self.datastructure
        precision = self.float_precision

        tags_formatter = {"xyz": gkey_format,
                          "key-xyz": key_gkey_format}

        if text_tag not in tags_formatter:
            return None

        text_labels = {}
        formatter = tags_formatter[text_tag]

        for node in ds.nodes():
            label = formatter(node)
            text_labels[node] = label

        return text_labels

    def _text_labels_edges(self, text_tag):
        """
        Generates text labels to plot on the edges of a ``FormDiagram``.

        Input
        -----
        text_tag : `str`
            Tag query.
            Supported tags are: "force", "length", and "force-length".

        Returns
        -------
        text_labels : ``list``
            A list of text labels
        """
        def force_format(x):
            return "{0:.{1}}".format(ds.edge_force(x), precision)

        def length_format(x):
            return "{0:.{1}}".format(ds.edge_length(*x), precision)

        def force_length_format(x):
            return "f: {} / lt: {}".format(force_format(x), length_format(x))

        ds = self.datastructure
        precision = self.float_precision

        tags_formatter = {"force": force_format,
                          "length": length_format,
                          "force-length": force_length_format}

        if text_tag not in tags_formatter:
            return None

        text_labels = {}
        formatter = tags_formatter[text_tag]

        for edge in ds.edges():
            label = formatter(edge)
            text_labels[edge] = label

        return text_labels


if __name__ == "__main__":
    pass



from compas.datastructures import Network

from compas.geometry import length_vector

from compas_cem.diagrams import NodeMixins
from compas_cem.diagrams import EdgeMixins


__all__ = ["Diagram"]


# ==============================================================================
# Diagram
# ==============================================================================


class Diagram(NodeMixins, EdgeMixins, Network):
    """
    Base class that shares functionality across diagrams.
    """
    def __init__(self, *args, **kwargs):
        super(Diagram, self).__init__(*args, **kwargs)

        self.update_default_node_attributes({"x": 0.0,
                                             "y": 0.0,
                                             "z": 0.0,
                                             "qx": 0.0,
                                             "qy": 0.0,
                                             "qz": 0.0,
                                             "rx": 0.0,
                                             "ry": 0.0,
                                             "rz": 0.0,
                                             "_k": None,
                                             "type": None})

        self.update_default_edge_attributes({"type": None,
                                             "length": 0.0,
                                             "force": 0.0})

        self.attributes["gkey_node"] = {}
        self.attributes["tol"] = "3f"

# ==============================================================================
# Properties
# ==============================================================================

    @property
    def tol(self):
        """
        The floating point tolerance for the node coordinates of the diagram.
        Defaults to 0.001.
        """
        return self.attributes["tol"]

    @tol.setter
    def tol(self, tol):
        """
        """
        self.attributes["tol"] = tol

    @property
    def gkey_node(self):
        """
        A dictionary that maps geometric keys to node keys.
        """
        return self.attributes["gkey_node"]

# ==============================================================================
#  Node collections
# ==============================================================================

    def support_nodes(self):
        """
        Nodes whose position is meant to be fixed in space.

        Yields
        -------
        support_node : ``int``
            The key of the next support node.
        """
        return self.nodes_where({"type": "support"})

    def loaded_nodes(self, min_force=1e-6):
        """
        Iterates over all the nodes with a large-enough load applied.

        Parameters
        ----------
        min_force : ``float``
            The minimum force magnitude to consider a node loaded.
            Defaults to ``1e-6``.

        Yields
        -------
        loaded_node : ``int``
            The key of the next loaded node.
        """
        for node in self.nodes():
            if self.is_node_loaded(node, min_force):
                yield node

# ==============================================================================
# Node Filters
# ==============================================================================

    def is_node_support(self, node):
        """
        Checks if a node is a support.

        Parameters
        ----------
        node : ``int``
            A node key.

        Returns
        -------
        flag : ``bool``
            ``True``if the node is a support. ``False`` otherwise.
        """
        return self.node_attribute(key=node, name="type") == "support"

    def is_node_loaded(self, node, min_force=1e-6):
        """
        Checks if there is a large-enough load applied to a node.

        Parameters
        ----------
        node : ``int``
            A node key.
        min_force : ``float``
            The minimum force magnitude to consider a node loaded.
            Defaults to ``1e-6``.
        Returns
        -------
        flag : ``bool``
            ``True``if the node is a support. ``False`` otherwise.
        """
        return length_vector(self.node_load(node)) > min_force

# ==============================================================================
# Node Attributes
# ==============================================================================

    def node_load(self, node):
        """
        Gets the load applied at a node.

        Parameters
        ----------
        node : ``int``
            A node key.

        Returns
        -------
        load_vector: ``list``
            A vector with xyz components of the load.
        """
        return self.node_attributes(key=node, names=["qx", "qy", "qz"])

    def reaction_force(self, node):
        """
        Gets the reaction force vector at a node support.

        Parameters
        ----------
        node : ``int``
            A node key.

        Returns
        -------
        type : ``list``
            The residual force vector.
        """
        return self.node_attributes(key=node, names=["rx", "ry", "rz"])

# ==============================================================================
# Edge Attributes
# ==============================================================================

    def edge_force(self, edge):
        """
        Gets the force value at an edge.

        Parameters
        ----------
        edge : ``tuple``
            The u, v edge key.

        Return
        ------
        force : ``float``
            The force value in the edge.
        """
        return self.edge_attribute(key=edge, name="force")

    def edge_length_2(self, edge):
        """
        Gets the length of an edge.

        Parameters
        ----------
        edge : ``tuple``
            The u, v edge key.

        Return
        ------
        length : ``float``
            The edge length.
        """
        return self.edge_attribute(key=edge, name="length")

# ==============================================================================
# Edge Vectors
# ==============================================================================

#     def incoming_edge_vectors(self, node, edges, normalize=False):
#         """
#         Calculates the edge vectors so that they point towards a node.

#         Parameters
#         ----------
#         node : ``int``
#             A node key.
#         edges : ``list``
#             A list of edge keys.
#         normalize : ``bool``
#             A boolean flag to normalize all the resulting edge vectors.
#             Defaults to ``False``.

#         Returns
#         -------
#         edge_vectors : ``list``
#             A list of xyz vectors.
#         """
#         vectors = []
#         for u, v in edges:
#             other_node = u if u != node else v
#             vector = self.two_node_vector((other_node, node), normalize)

#             if normalize:
#                 vector = normalize_vector(vector)

#             vectors.append(vector)

#         return vectors


#     def two_node_vector(self, nodes, normalize=False):
#         """
#         Calculates the vector between two nodes in a diagram.

#         Parameters
#         ----------
#         nodes : ``list``
#             A list with two node keys.
#         normalize : ``bool``
#             A boolean flag to normalize all the resulting edge vectors.
#             Defaults to ``False``.

#         Returns
#         -------
#         vector : ``list``
#             The calculated xyz vector.
#         """
#         assert len(nodes) == 2, "Supply only two nodes!"
#         vector = subtract_vectors(*[self.node_coordinates(n) for n in nodes])
#         if not normalize:
#             return vector
#         return normalize_vector(vector)

# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":
    pass



import compas

if compas.RHINO:
    from compas_rhino.geometry import RhinoPoint


__all__ = ["Node"]

# ==============================================================================
# Node
# ==============================================================================


class Node(object):
    """
    A node.
    """
    def __init__(self, key=None, xyz=[0.0, 0.0, 0.0]):
        self.key = key
        self.xyz = xyz
        self.attributes = {}

    @classmethod
    def from_point(cls, point, *args, **kwargs):
        """
        Create a node from a point described by its xyz coordinates.

        Parameters
        ----------
        point : ``tuple`` or ``list``
            The xyz coordinates of the point.
        *args : ``tuple``
            Additional arguments.
        **kwargs : ``dict``
            Extra keyword arguments.

        Returns
        -------
        node : ``Node``
            A node object.
        """
        return cls(xyz=point)

    @classmethod
    def from_rhino_point(cls, rhino_point, *args, **kwargs):
        """
        Create a node from a rhino point.

        Parameters
        ----------
        rhino_point : ``Rhino.Geometry.Point3d``
            A rhino point.
        *args : ``tuple``
            Additional arguments.
        **kwargs : ``dict``
            Extra keyword arguments.

        Returns
        -------
        node : ``Node``
            A node object.
        """
        point = RhinoPoint.from_geometry(rhino_point).to_compas()
        return cls.from_point(point)

    def __repr__(self):
        """
        """
        return "{}(key={}, xyz={})".format(self.__class__.__name__, self.key, self.xyz)


if __name__ == "__main__":
    pass



import compas

if compas.RHINO:
    from compas_rhino.geometry import RhinoPoint
    from compas_rhino.geometry import RhinoVector


__all__ = ["NodeLoad"]

# ==============================================================================
# Node Load
# ==============================================================================


class NodeLoad(object):
    """
    A load defined by a position and a vector.

    Parameters
    ----------
    node : ``int``
        A node key
    vector : ``list`` of ``float``
        The load magnitude of the point load in xyz directions.
    """
    def __init__(self, node, vector=[0, 0, -1]):
        self.node = node
        self.vector = vector
        self.xyz = None

    @classmethod
    def from_point_and_vector(cls, point, vector):
        """
        Create a NodeLoad from a point and a vector.

        Parameters
        ----------
        point : ``list`` of ``float``
            The xyz coordinates of the positions where to applied the load.
        vector : ``list`` of ``float``
            The xyz coordinates of the load vector.

        Returns
        -------
        load : ``NodeLoad``
            A load object.
        """
        load = cls(node=None, vector=vector)
        load.xyz = point
        return load

    @classmethod
    def from_rhino_point_and_vector(cls, rhino_point, rhino_vector):
        """
        Create a NodeLoad from a rhino point and a rhino vector.

        Parameters
        ----------
        rhino_point : ``Rhino.Geometry.Point``
            The load application point.
        rhino_vector : ``Rhino.Geometry.Vector3d``
            A vector that encodes the magnitude and direction of the load.

        Returns
        -------
        load : ``NodeLoad``
            A load object.
        """
        point = RhinoPoint.from_geometry(rhino_point).to_compas()
        vector = RhinoVector.from_geometry(rhino_vector).to_compas()
        return cls.from_point_and_vector(point, vector)


if __name__ == "__main__":
    pass



from compas_cem.elements import Edge


class TrailEdge(Edge):
    """
    A trail edge.
    """
    def __init__(self, u, v, length):
        super(TrailEdge, self).__init__(u, v)
        self.attributes = {"length": length, "type": "trail"}

    def __repr__(self):
        """
        """
        return "{}(length={})".format(self.__class__.__name__, self.attributes["length"])


if __name__ == "__main__":
    pass



from compas_cem.optimization.constraint import Constraint

from compas.geometry import distance_point_point_sqrd

import numpy as np


__all__ = ["TrimeshConstraint"]


class TrimeshConstraint(Constraint):
    """
    Pulls the xyz position of a node to a target triangular mesh.

    TODO: trimesh complaints with autograd boxes when parsing into array.
    """
    def __init__(self, node=None, trimesh=None, weight=1.0):
        super(TrimeshConstraint, self).__init__(node, trimesh, weight)

    def error(self, data):
        """
        """
        point_a = self.reference(data)
        point_b = self.target(point_a)

        return distance_point_point_sqrd(point_a, point_b) * self.weight

    def reference(self, data):
        """
        """
        return data["node_xyz"][self.key()]

    def target(self, ref):
        """
        """
        points = np.array([ref])
        trimesh = self._target
        closest, dist, _ = trimesh.nearest.on_surface(points)

        return closest.tolist().pop()


if __name__ == "__main__":
    pass



from compas.geometry import closest_point_on_line
from compas.geometry import distance_point_point_sqrd

from compas_cem.optimization.constraints import Constraint


__all__ = ["LineConstraint"]


class LineConstraint(Constraint):
    """
    Pulls the xyz position of a node to a target line ray.
    """
    def __init__(self, node=None, line=None, weight=1.0):
        super(LineConstraint, self).__init__(node, line, weight)

    def error(self, data):
        """
        The error between the xyz coords of a node and its closest point on a line.

        Returns
        -------
        error : ``float``
            The squared distance between the two points.
        """
        point_a = self.reference(data)
        point_b = self.target(point_a)

        return distance_point_point_sqrd(point_a, point_b) * self.weight

    def reference(self, data):
        """
        """
        point = data["node_xyz"][self.key()]

        return point

    def target(self, point):
        """
        """
        line = self._target

        return closest_point_on_line(point, line)


if __name__ == "__main__":

    from time import time

    from compas.geometry import Line
    from compas.geometry import Point

    from compas_cem.diagrams import TopologyDiagram

    from compas_cem.elements import Node
    from compas_cem.elements import TrailEdge
    from compas_cem.elements import DeviationEdge

    from compas_cem.loads import NodeLoad

    from compas_cem.supports import NodeSupport

    from compas_cem.optimization import Optimizer
    from compas_cem.optimization import DeviationEdgeParameter
    from compas_cem.optimization import TrailEdgeParameter
    from compas_cem.optimization import PointConstraint

    from compas_cem.equilibrium import static_equilibrium

    from compas_cem.plotters import FormPlotter

    # create a topology diagram
    topology = TopologyDiagram()

    # add nodes
    topology.add_node(Node(0, [0.0, 0.0, 0.0]))
    topology.add_node(Node(1, [1.0, 0.0, 0.0]))
    topology.add_node(Node(2, [2.0, 0.0, 0.0]))
    topology.add_node(Node(3, [3.0, 0.0, 0.0]))
    topology.add_node(Node(4, [4.0, 0.0, 0.0]))

    # add edges with negative values for a compression-only structure
    topology.add_edge(TrailEdge(0, 1, length=-1.0))
    topology.add_edge(DeviationEdge(1, 2, force=-1.0))
    topology.add_edge(TrailEdge(2, 3, length=-1.0))
    topology.add_edge(TrailEdge(3, 4, length=-1.0))

    # add supports
    topology.add_support(NodeSupport(0))
    topology.add_support(NodeSupport(4))

    # add loads
    topology.add_load(NodeLoad(1, [0.0, -1.0, 0.0]))
    topology.add_load(NodeLoad(2, [0.0, -1.0, 0.0]))

    # calculate equilibrium
    topology.build_trails()
    form = static_equilibrium(topology)

    # optimization
    optimizer = Optimizer()

    optimizer.add_parameter(DeviationEdgeParameter((1, 2), 1.0, 1.0))
    optimizer.add_parameter(TrailEdgeParameter((2, 3), 1.0, 1.0))
    optimizer.add_parameter(TrailEdgeParameter((3, 4), 1.0, 1.0))

    point_a = Point(3.0, -0.707, 0.0)
    optimizer.add_constraint((PointConstraint(3, point_a)))

    line = Line([4.0, 0.0, 0.0], [4.0, 1.0, 0.0])
    optimizer.add_constraint((LineConstraint(4, line)))

    # optimization settings
    start = time()
    algo = "LD_LBFGS"  # LN_BOBYQA, LD_LBFGS, LD_MMA
    iters = 100  # 100
    eps = 1e-6  # 1e-4

    # optimize
    form = optimizer.solve_nlopt(topology, algo, iters, eps)

    # print out results
    print("Elapsed time: {}".format(time() - start))
    print("Total error: {}".format(optimizer.penalty))

    # plot
    plotter = FormPlotter(form, figsize=(16, 9))

    plotter.draw_nodes(radius=0.03, text="key-xyz")
    plotter.draw_edges(text="force-length")
    plotter.draw_loads(scale=0.25)
    plotter.draw_reactions(scale=0.10)
    plotter.show()



from compas.utilities import iterable_like


__all__ = ["form_plotter_proxy",
           "topology_plotter_proxy"]


def form_plotter_proxy(**kwargs):
    """
    """
    from compas_cem.plotters import FormPlotter

    # unpack kwargs
    form = kwargs["form"]
    filepath = kwargs["filepath"]

    edge_key = kwargs.get("edge_key")
    edge_width = kwargs.get("edge_width", 1.0)

    node_key = kwargs.get("node_key")
    node_radius = kwargs.get("node_radius", 0.1)
    node_edgewidth = kwargs.get("node_edgewidth", 1.0)

    load_width = kwargs.get("load_width", 1.0)
    load_scale = kwargs.get("load_scale", 1.0)
    load_gap = kwargs.get("load_gap", 0.1)

    reaction_width = kwargs.get("reaction_width", 1.0)
    reaction_scale = kwargs.get("reaction_scale", 1.0)
    reaction_gap = kwargs.get("reaction_gap", 0.1)

    segment = kwargs.get("segment", [])
    segment_color = kwargs.get("segment_color") or iterable_like(segment, [], (50, 50, 50))
    segment_width = kwargs.get("segment_width") or iterable_like(segment, [], 0.5)
    segment_ls = kwargs.get("segment_ls") or iterable_like(segment, [], "--")

    fig_width = kwargs.get("fig_width", 16)
    fig_height = kwargs.get("fig_height", 9)
    fig_dpi = kwargs.get("fig_dpi", 100)

    frame_polygon = kwargs.get("frame_polygon", None)

    # plot
    plotter = FormPlotter(form, figsize=(fig_width, fig_height), dpi=fig_dpi)

    plotter.draw_edges(keys=edge_key, width=edge_width)
    plotter.draw_nodes(keys=node_key, radius=node_radius, edgewidth=node_edgewidth)
    plotter.draw_loads(width=load_width, scale=load_scale, gap=load_gap)
    plotter.draw_residuals(width=reaction_width, scale=reaction_scale, gap=reaction_gap)

    if frame_polygon:
        polygons = [{"points": frame_polygon, "edgecolor": (255, 255, 255)}]
        plotter.draw_polygons(polygons)

    for seg, ls, color, width in zip(segment, segment_ls, segment_color, segment_width):
        if sum(color) == 255 * 3:
            continue
        plotter.draw_segments([seg], color, width, ls)

    plotter.save(filepath)


def topology_plotter_proxy(**kwargs):
    """
    """
    from compas_cem.plotters import TopologyPlotter

    # unpack kwargs
    topology = kwargs["topology"]
    filepath = kwargs["filepath"]

    edge_key = kwargs.get("edge_key")
    edge_width = kwargs.get("edge_width", 1.0)

    node_key = kwargs.get("node_key")
    node_radius = kwargs.get("node_radius", 0.1)
    node_edgewidth = kwargs.get("node_edgewidth", 1.0)

    load_radius = kwargs.get("load_radius", 1.0)
    load_width = kwargs.get("load_width", 1.0)

    segment = kwargs.get("segment", [])
    segment_color = kwargs.get("segment_color") or iterable_like(segment, [], (50, 50, 50))
    segment_width = kwargs.get("segment_width") or iterable_like(segment, [], 0.5)
    segment_ls = kwargs.get("segment_ls") or iterable_like(segment, [], "--")

    fig_width = kwargs.get("fig_width", 16)
    fig_height = kwargs.get("fig_height", 9)
    fig_dpi = kwargs.get("fig_dpi", 100)

    frame_polygon = kwargs.get("frame_polygon", None)

    # plot
    plotter = TopologyPlotter(topology, figsize=(fig_width, fig_height), dpi=fig_dpi)

    plotter.draw_loads(keys=node_key, radius=load_radius, width=load_width)
    plotter.draw_nodes(keys=node_key, radius=node_radius, edgewidth=node_edgewidth)
    plotter.draw_edges(keys=edge_key, width=edge_width)

    if frame_polygon:
        polygons = [{"points": frame_polygon, "edgecolor": (255, 255, 255)}]
        plotter.draw_polygons(polygons)

    for seg, ls, color, width in zip(segment, segment_ls, segment_color, segment_width):
        if sum(color) == 255 * 3:
            continue
        plotter.draw_segments([seg], color, width, ls)

    plotter.save(filepath)



import json

from compas.base import Base


__all__ = ["Serializable", "cls_from_dtype"]


# ------------------------------------------------------------------------------
# Base Class
# ------------------------------------------------------------------------------


class Serializable(Base):

    # --------------------------------------------------------------------------
    # Data
    # --------------------------------------------------------------------------

    @property
    def data(self):
        """
        """
        return

    @data.setter
    def data(self, data):
        """
        """
        return

    # --------------------------------------------------------------------------
    # IO - Data
    # --------------------------------------------------------------------------

    @classmethod
    def from_data(cls, data):
        """
        Creates a serializable object from structured data.

        Parameters
        ----------
        data : ``dict``
            The data dictionary.

        Returns
        -------
        object : ``cls``
            An instance of a ``cls`` object.

        Notes
        -----
        This constructor  is meant to be used in conjunction with the
        corresponding ``to_data()`` method.
        """
        serializable = cls()
        serializable.data = data
        return serializable

    def to_data(self):
        """
        Returns a dictionary of structured data representing a serializable object.

        Returns
        -------
        data : ``dict``
            The structured data.

        Notes
        ----
        This method produces the data that can be used in conjunction with the
        corresponding *from_data()* method.
        """
        return self.data

    # --------------------------------------------------------------------------
    # IO - JSON
    # --------------------------------------------------------------------------

    @classmethod
    def from_json(cls, filepath):
        """
        Construct a serializable object from structured data contained in a json file.

        Parameters
        ----------
        filepath : str
            The path to the json file.

        Returns
        -------
        object : ``cls``
            An instance of a ``cls`` object.

        Notes
        -----
        This constructor method is meant to be used in conjunction with the
        corresponding *to_json()* method.
        """
        with open(filepath, 'r') as fp:
            data = json.load(fp)
        optimizer = cls()
        optimizer.data = data
        return optimizer

    def to_json(self, filepath, pretty=False):
        """
        Serialize this object to a structured json.

        Parameters
        ----------
        filepath : str
            The path to the json file.

        Notes
        -----
        This constructor method is meant to be used in conjunction with the
        corresponding *from_json()* method.
        """
        with open(filepath, 'w+') as f:
            if pretty:
                json.dump(self.data, f, sort_keys=True, indent=4)
            else:
                json.dump(self.data, f)
        return

    def datatype(self):
        """
        Get the data type of this COMPAS object.

        Returns
        -------
        datatype : ``str``
            The data type of the COMPAS object in the following format:
            '{}/{}'.format(o.__class__.__module__, o.__class__.__name__).
        """
        return self.object_datatype(self)

    @staticmethod
    def object_datatype(o):
        return "{}/{}".format(o.__class__.__module__, o.__class__.__name__)

    @staticmethod
    def object_cls_from_dtype(dtype):
        return cls_from_dtype(dtype)

# ------------------------------------------------------------------------------
# Class from datatype
# ------------------------------------------------------------------------------


def cls_from_dtype(dtype):
    """
    Get the class object corresponding to a COMPAS data type specification.

    Parameters
    ----------
    dtype : str
        The data type of the COMPAS object in the following format:
        '{}/{}'.format(o.__class__.__module__, o.__class__.__name__).

    Returns
    -------
    :class:`compas.base.Base`

    Raises
    ------
    ValueError
        If the data type is not in the correct format.
    ImportError
        If the module can't be imported.
    AttributeError
        If the module doesn't contain the specified data type.
    """
    mod_name, attr_name = dtype.split('/')
    module = __import__(mod_name, fromlist=[attr_name])
    return getattr(module, attr_name)

# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------


if __name__ == "__main__":
    pass



from compas_cem.elements import Node


__all__ = [
    "EdgeMixins"
]

# ==============================================================================
# Edge Mixins
# ==============================================================================


class EdgeMixins(object):
    """
    """
    def add_edge(self, edge):
        """
        Adds a deviation or a trail edge.

        Parameters
        ----------
        edge : ``Edge``
            An edge element.

        Returns
        -------
        key : ``tuple``
            An edge key.
        """
        edge_keys = []

        for node in edge:
            key = self.node_key(node)

            if key is not None:
                edge_keys.append(key)
                continue

            xyz = None
            if not isinstance(key, int):
                xyz = node

            key = self.add_node(Node(key, xyz))
            edge_keys.append(key)

        u, v = edge_keys
        attr = {k: v for k, v in edge.attributes.items()}
        return super(EdgeMixins, self).add_edge(u=u, v=v, attr_dict=attr)

# ==============================================================================
# Main
# ==============================================================================


if __name__ == "__main__":
    from compas_cem.diagrams import TopologyDiagram
    from compas_cem.elements import DeviationEdge
    from compas.geometry import Line

    topology = TopologyDiagram()
    topology.add_node(Node.from_point([0.0, 1.0, 0.0]))

    line = Line([0.0, 1.0, 0.0], [2.0, 0.0, 0.0])
    edge = topology.add_edge(DeviationEdge.from_line(line, 10))

    assert edge == (0, 1), "{}".format(edge)
    assert topology.node_xyz(0) == [0.0, 1.0, 0.0]
    assert topology.node_xyz(1) == [2.0, 0.0, 0.0]
    assert topology.edge_force(edge) == 10.0
    assert topology.number_of_nodes() == 2
    assert topology.number_of_edges() == 1



from abc import abstractmethod

from compas_cem.optimization import Serializable


__all__ = ["Constraint"]


class Constraint(Serializable):
    """
    The blueprint of a constraint.
    """
    def __init__(self, key, target, weight):
        self._key = key  # a topological key
        self._target = target  # a geometric target
        self._weight = weight  # the strength of the constraint

    def key(self):
        """
        The key to an edge or a node in a form diagram.

        Returns
        -------
        key : ``int`` or ``tuple``
            The key to a node of an edge.
        """
        return self._key

    def target(self):
        """
        The target to reach.

        Returns
        -------
        target : ``object``
            A target object.
        """
        return self._target

    @abstractmethod
    def error(self):
        """
        Calculate the error to the target.
        """
        return

    @property
    def weight(self):
        """
        The strength of the constraint.
        """
        return self._weight

# ------------------------------------------------------------------------------
# Data
# ------------------------------------------------------------------------------
    # TODO: Update serialization after autodiff. Danger of not working in Rhino!

    @property
    def data(self):
        """
        A data dictionary that represents a ``Goal`` object.

        Returns
        -------
        data : ``dict``
            A dictionary that contains the following key-value pairs:
        """

        data = {}
        data["datatype"] = self.datatype()
        data["node_key"] = str(self._key)
        data["target_datatype"] = self.object_datatype(self._target)
        data["target"] = self._target.to_data()

        return data

    @data.setter
    def data(self, data):
        """
        Overwrites this object's attributes with a data dictionary.

        Parameters
        ----------
        data : ``dict``
            A data dictionary.
        """
        self._key = int(data["node_key"])
        target_cls = self.object_cls_from_dtype(data["target_datatype"])
        target = target_cls.from_data(data["target"])
        self._target = target


if __name__ == "__main__":
    pass



from ast import literal_eval
from math import fabs

from compas_cem.optimization import Serializable


__all__ = ["TrailEdgeParameter",
           "DeviationEdgeParameter"]

# ------------------------------------------------------------------------------
# Base Edge Parameter
# ------------------------------------------------------------------------------


class EdgeParameter(Serializable):
    """
    Parametrize and edge to solve an optimization problem.
    """
    def __init__(self, key, bound_low, bound_up):
        self._key = key
        self._bound_up = bound_up
        self._bound_low = bound_low
        self._attr_name = None

    def key(self):
        """
        """
        return self._key

    def start_value(self, topology):
        """
        """
        val = topology.edge_attribute(key=self.key(), name=self._attr_name)
        return val

    def bound_low(self, topology):
        """
        """
        return self.start_value(topology) - fabs(self._bound_low)

    def bound_up(self, topology):
        """
        """
        return self.start_value(topology) + fabs(self._bound_up)

    def attr_name(self):
        """
        """
        return self._attr_name

# ------------------------------------------------------------------------------
# Data
# ------------------------------------------------------------------------------

    @property
    def data(self):
        """
        A data dictionary that represents an ``EdgeParameter`` object.

        Returns
        -------
            data : ``dict``
            A dictionary that contains the following key-value pairs:

            * "key" : ``tuple``
            * "bound up" : ``float``
            * "bound low" : ``float``
            * "attr name" : ``str``

        Notes
        -----
        All dictionary keys are converted to their representation value
        (``repr(key)``) to ensure compatibility of all allowed key types with
        the JSON serialization format, which only allows for dict keys that are strings.
        """

        data = {}

        data["key"] = repr(self._key)
        data["bound_up"] = self._bound_up
        data["bound_low"] = self._bound_low
        data["attr_name"] = self._attr_name
        data["datatype"] = self.datatype()

        return data

    @data.setter
    def data(self, data):
        """
        Overwrites this object's attributes with a data dictionary.

        Parameters
        ----------
        data : ``dict``
            A data dictionary.
        """
        self._key = tuple(literal_eval(data["key"]))
        self._bound_up = float(data["bound_up"])
        self._bound_low = float(data["bound_low"])
        self._attr_name = str(data["attr_name"])

# ------------------------------------------------------------------------------
# Trail Edge Parameter
# ------------------------------------------------------------------------------


class TrailEdgeParameter(EdgeParameter):
    """
    Sets the length of a trail edge as an optimization parameter.
    """
    def __init__(self, key=None, bound_low=None, bound_up=None):
        super(TrailEdgeParameter, self).__init__(key, bound_low, bound_up)
        self._attr_name = "length"

# ------------------------------------------------------------------------------
# Deviation Edge Parameter
# ------------------------------------------------------------------------------


class DeviationEdgeParameter(EdgeParameter):
    """
    Sets the force of a deviation edge as an optimization parameter.
    """
    def __init__(self, key=None, bound_low=None, bound_up=None):
        super(DeviationEdgeParameter, self).__init__(key, bound_low, bound_up)
        self._attr_name = "force"

# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------


if __name__ == "__main__":
    pass



"""
compas_cem.equilibrium
****************************

.. currentmodule:: compas_cem.equilibrium


Equilibrium
============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    static_equilibrium
    static_equilibrium_numpy

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


# from .<module> import *
from .force import *  # noqa F403

import compas
if not compas.IPY:
    from .force_numpy import *  # noqa F403

__all__ = [name for name in dir() if not name.startswith('_')]



from nlopt import opt

from nlopt import LD_AUGLAG
from nlopt import LD_LBFGS
from nlopt import LD_MMA
from nlopt import LD_SLSQP
from nlopt import LD_TNEWTON

from nlopt import LN_BOBYQA
from nlopt import LN_COBYLA
from nlopt import LN_SBPLX


__all__ = ["nlopt_algorithm",
           "nlopt_algorithms",
           "nlopt_solver"]


def nlopt_algorithm(name):
    """
    Fetches an optimization algorithm from the nlopt library by name.

    Parameters
    ----------
    name : ``str``
        The name of the algorithm to search for.

    Returns
    -------
    algorithm : ``nlopt.algorithm``
        An nlopt algorithm object.

    Notes
    -----
    Only the following local optimization algorithms are supported:

    Gradient-based:

        - LD_MMA
        - LD_LBFGS
        - LD_AUGLAG
        - LD_SLSQP
        - LD_TNEWTON

    Derivative-free:

        - LN_BOBYQA
        - LN_COBYLA
        - LN_SBPLX

    Refer to nlopt's docs for more details on their theoretical underpinnings.
    """

    algorithms = nlopt_algorithms()
    return algorithms[name]


def nlopt_algorithms():
    """
    A dictionary with all the supported nlopt algorithms.

    Returns
    -------
    algorithms : ``dict``
        A dictionary that maps algorithm names to nlopt algorithm objects.
    """
    algorithms = {}

    gradient_based = {
        "LD_MMA": LD_MMA,  # Method of moving asymptotes
        "LD_LBFGS": LD_LBFGS,  # Low-storage BFGS
        "LD_AUGLAG": LD_AUGLAG,  # Augmented lagrangian algorithm
        "LD_SLSQP": LD_SLSQP,  # Sequential quadratic programming algorithm
        "LD_TNEWTON": LD_TNEWTON,  # Preconditioned truncated Newton
    }

    derivative_free = {
        "LN_BOBYQA": LN_BOBYQA,  # BOBYQA
        "LN_COBYLA": LN_COBYLA,  # Constrained optimization by linear approx
        "LN_SBPLX": LN_SBPLX,  # Tom Rowan's "Subplex" algorithm
    }

    algorithms.update(derivative_free)
    algorithms.update(gradient_based)

    return algorithms


def nlopt_solver(f, algorithm, dims, bounds_up, bounds_low, iters, stopval, ftol):
    """
    Wrapper around a typical nlopt solver routine.
    """
    solver = opt(nlopt_algorithm(algorithm), dims)

    solver.set_lower_bounds(bounds_low)
    solver.set_upper_bounds(bounds_up)

    solver.set_maxeval(iters)

    if ftol is not None:
        solver.set_ftol_abs(ftol)  # abs per recommendation

    if stopval is not None:
        solver.set_stopval(stopval)

    solver.set_min_objective(f)

    return solver



"""
compas_cem.elements
****************************

.. currentmodule:: compas_cem.elements


Edges
======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    TrailEdge
    DeviationEdge

Node
=====

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Node
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


# from .<module> import *
from .node import *  # noqa F403
from .edge import *  # noqa F403
from .trail import *  # noqa F403
from .deviation import *  # noqa F403

__all__ = [name for name in dir() if not name.startswith('_')]



"""
compas_cem.plotters
****************************

.. currentmodule:: compas_cem.plotters


Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    FormPlotter
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


# from .<module> import *
import compas

if not compas.IPY:
    from .form_plotter import *  # noqa F403
    from .topology_plotter import *  # noqa F403

from .proxy import *  # noqa F403


__all__ = [name for name in dir() if not name.startswith('_')]



from compas_cem.optimization.constraints import Constraint

from compas.geometry import distance_point_point_sqrd


__all__ = ["PointConstraint"]


class PointConstraint(Constraint):
    """
    Pulls the xyz position of a node to a target point.
    """
    def __init__(self, node=None, point=None, weight=1.0):
        super(PointConstraint, self).__init__(node, point, weight)

    def error(self, data):
        """
        The error between the xyz coordinates of a node and a target point.

        Returns
        -------
        error : ``float``
            The squared distance between two points.
        """
        a = self.reference(data)
        b = self.target()

        return distance_point_point_sqrd(a, b) * self.weight

    def reference(self, data):
        """
        Gets the reference to compare the target against.
        """
        a = data["node_xyz"][self.key()]
        return a


if __name__ == "__main__":
    pass



import autograd.numpy as np

from compas_cem.diagrams import FormDiagram


__all__ = ["static_equilibrium_numpy"]


def static_equilibrium_numpy(topology, tmax=100, eta=1e-5, verbose=False, callback=None):
    """
    Generate a form diagram in static equilibrium using a numpy backend.

    Parameters
    ----------
    topology : ``TopologyDiagram``
        A valid topology diagram
    tmax : ``int``. Optional.
        Maximum number of iterations the algorithm will run for.
        Defaults to ``100``.
    eta: ``float``. Optional.
        Distance threshold that marks equilibrium convergence.
        This threshold is compared against the sum of distances of the nodes'
        positions from one iteration to the next one.
        If ``eta`` is hit before consuming ``kmax`` iterations, calculations
        will stop early. Defaults to ``1e-5``.
    verbose : ``bool``. Optional.
        Flag to print out internal operations. Defaults to ``False``.
    callback : ``function``. Optional.
        An optional callback function to run at every iteration.
    """
    attrs = equilibrium_state_numpy(topology, tmax, eta, verbose, callback)
    form = FormDiagram.from_topology_diagram(topology)
    form_update(form, **attrs)
    return form


def equilibrium_state_numpy(topology, tmax=100, eta=1e-5, verbose=False, callback=None):
    """
    Equilibrate forces in a topology.
    """
    trails = topology.trails()  # calls attribute self.attributes["trails"]

    # input, output
    node_xyz = {n: np.array(topology.node_coordinates(n)) for n in topology.nodes()}

    # output, mutable
    edge_forces = {e: np.array(topology.edge_force(e)) for e in topology.edges()}
    edge_lengths = {e: np.array(topology.edge_attribute(e, "length")) for e in topology.edges()}

    # input, immutable
    # numpy
    node_loads = {n: np.array(topology.node_load(n)) for n in topology.nodes()}
    # no numpy
    node_direct = {n: topology._connected_direct_deviation_edges(n) for n in topology.nodes()}
    node_indirect = {n: topology._connected_indirect_deviation_edges(n) for n in topology.nodes()}
    edge_keys = {e for e in topology.edges()}

    # internals
    trail_vectors = {}

    # output
    reaction_forces = {}
    trail_forces = {}

    for t in range(tmax):  # max iterations

        # store last positions for residual
        last_positions = {k: v for k, v in node_xyz.items()}

        for i in topology.sequences():  # layers

            for _, trail in trails.items():

                # if index is larger than available nodes in trails
                if i > (len(trail) - 1):
                    continue

                # get node key from trail
                node = trail[i]

                # set initial trail vector and position for first iteration
                pos = node_xyz[node]
                if i == 0:
                    t_vec = np.zeros(3)

                # otherwise, select last trail vector and position from dictionary
                else:
                    t_vec = trail_vectors[node]

                # node load
                q_vec = node_loads[node]

                # direct deviation edges vector
                dir_edges = node_direct[node]
                rd_vec = deviation_edges_resultant_vector(node, node_xyz, dir_edges, edge_forces)

                # indirect deviation edges vector
                ri_vec = np.zeros(3)
                if t > 0:
                    indir_edges = node_indirect[node]
                    ri_vec = deviation_edges_resultant_vector(node, node_xyz, indir_edges, edge_forces)

                # node equilibrium, bottleneck 60%
                t_vec = node_equilibrium(t_vec, q_vec, rd_vec, ri_vec)

                # if this is the last node, exit
                if i == (len(trail) - 1):
                    continue

                # otherwise, pick next node in the trail
                next_node = trail[i + 1]

                # correct edge key
                edge = (node, next_node)
                if edge not in edge_keys:
                    edge = (next_node, node)

                # query trail edge's length
                length = edge_lengths[edge]

                trail_force = length_vector(t_vec)
                trail_forces[edge] = trail_force

                next_pos = pos + length * t_vec / trail_force

                # store new position and trail vector
                trail_vectors[next_node] = t_vec

                # store node coordinates
                node_xyz[next_node] = next_pos

                # store reaction force in support node
                reaction_forces[next_node] = t_vec

                # do callback
                if callback:
                    callback()

        # if this is the first iteration, move directly to the next one
        if t == 0:
            continue

        # TODO: simplify by iterating only over values?
        pos_array = np.array([pos for key, pos in node_xyz.items()])
        last_pos_array = np.array([last_positions[key] for key, pos in node_xyz.items()])

        # calculate residual
        residual = np.sqrt(np.sum(np.square(last_pos_array - pos_array)))
        # if residual smaller than threshold, stop iterating
        if residual < eta:
            break

    # if residual larger than threshold after kmax iterations, raise error
    if residual > eta:
        raise ValueError("Over {} iters. residual: {} > eta: {}".format(tmax, residual, eta))

    # print log
    if verbose:
        msg = "====== Completed Equilibrium in {} iters. Residual: {}======"
        print(msg.format(t, residual))

    eq_state = {}
    eq_state["node_xyz"] = node_xyz
    eq_state["trail_forces"] = trail_forces
    eq_state["reaction_forces"] = reaction_forces

    # return node_xyz, trail_forces, reaction_forces
    return eq_state


def form_update(form, node_xyz, trail_forces, reaction_forces):
    """
    Update the node and edge attributes of a form after equilibrating it.
    """
    # assign nodes' coordinates
    for node, xyz in node_xyz.items():
        form.node_attributes(key=node, names=["x", "y", "z"], values=xyz)

    # assign forces on trail edges
    for edge, tforce in trail_forces.items():
        tlength = form.edge_attribute(key=edge, name="length")
        tforce = np.copysign(tforce, tlength)
        form.edge_attribute(key=edge, name="force", value=tforce)

    # assign reaction forces
    for node in form.support_nodes():
        rforce = reaction_forces[node]
        form.node_attributes(key=node, names=["rx", "ry", "rz"], values=rforce)

    # assign lengths to deviation edges
    for u, v in form.edges():
        # length = form.edge_length(u, v)
        length = length_vector(node_xyz[u] - node_xyz[v])
        form.edge_attribute(key=(u, v), name="length", value=length)


def node_equilibrium(t_vec, q_vec, rd_vec, ri_vec):
    """
    Calculates the equilibrium of trail and deviation forces at a node.

    Parameters
    ----------
    t_vec : ``list``
        A trail vector.
    indirect : ``bool``
        Flag to consider indirect deviation edges in the calculation.
        Defaults to ``False``.

    Returns
    -------
    t_vec : ``list``
        The new trail vector.
    """
    tvec_in = -1.0 * t_vec

    return trail_vector_out(tvec_in, q_vec, rd_vec, ri_vec)  # bottleneck


def deviation_edges_resultant_vector(node, node_xyz, deviation_edges, edge_forces):
    """
    Adds up the force vectors of the deviation edges incident to a node.

    Parameters
    ----------
    topology : ``TopologyDiagram``
        A topology diagram.
    node : ``int``
        A node key.
    node_xyz : ``dict``
        A dictionary with node keys and xyz coordinates as values.
    deviation_edges : ``list``
        A list with deviation edges keys.

    Returns
    -------
    rvec : ``list``
        The resulting force vector.
    """
    r_vec = np.zeros(3)
    if not deviation_edges:
        return r_vec

    for edge in deviation_edges:
        force = edge_forces[edge]
        vector = incoming_edge_vector(node, node_xyz, edge, normalize=True)
        r_vec = r_vec + force * vector

    # bottleneck 40%
    # TODO: check if np.array forces broadcasts well with vectors array

    return r_vec


def direct_deviation_edges_resultant_vector(topology, node, node_xyz):
    """
    Adds up the force vectors of the direct deviation edges incident to a node.

    Parameters
    ----------
    topology : ``TopologyDiagram``
        A topology diagram.
    node : ``int``
        A node key.

    Returns
    -------
    rvec : ``list``
        The resulting force vector.
    """
    deviation_edges = topology._connected_direct_deviation_edges(node)
    res = deviation_edges_resultant_vector(topology, node, node_xyz, deviation_edges)

    return res


def indirect_deviation_edges_resultant_vector(topology, node, node_xyz):
    """
    Adds up the force vectors of the indirect deviation edges incident to a node.

    Parameters
    ----------
    topology : ``TopologyDiagram``
        A topology diagram.
    node : ``int``
        A node key.

    Returns
    -------
    rvec : ``list``
        The resulting force vector.
    """
    deviation_edges = topology._connected_indirect_deviation_edges(node)

    return deviation_edges_resultant_vector(topology, node, node_xyz, deviation_edges)


def trail_vector_out(tvec_in, q_vec, rd_vec, ri_vec):
    """
    Calculate an outgoing trail vector.

    Parameters
    ----------
    tvec_in : ``list``
        An incoming trail vector.
    q_vec : ``list``
        A load vector.
    rd_vec : ``list``
        The direct deviation edges resultant force vector.
    ri_vec : ``list``
        The indirect deviation edges resultant force vector.

    Returns
    -------
    tvec_out : ``list``
        An outgoing trail vector.
    """
    return -1.0 * (tvec_in + q_vec + rd_vec + ri_vec)

# ------------------------------------------------------------------------------
# Utilities
# ------------------------------------------------------------------------------


def incoming_edge_vectors(node, node_xyz, edges, normalize=False):
    """
    Temporary alternative to Diagram.incoming_edge_vectors()
    """
    vectors = []
    for u, v in edges:
        other = u if u != node else v
        vector = vector_two_nodes(node_xyz[other], node_xyz[node], normalize)
        vectors.append(vector)

    return vectors


def incoming_edge_vector(node, node_xyz, edge, normalize=False):
    """
    Temporary alternative to Diagram.incoming_edge_vectors()
    """
    u, v = edge
    other = u if u != node else v

    return vector_two_nodes(node_xyz[other], node_xyz[node], normalize)


def vector_two_nodes(a, b, normalize=False):
    """
    Calculates the vector between the xyz coordinates of two noddes.

    Parameters
    ----------
    a : ``list``
        The xyz coordinates of the first node
    b : ``list``
        The xyz coordinates of the second node
    normalize : ``bool``
        A boolean flag to normalize all the resulting edge vectors.
        Defaults to ``False``.

    Returns
    -------
    vector : ``list``
        The calculated xyz vector.
    """
    vector = a - b
    if not normalize:
        return vector
    return normalize_vector(vector)


def normalize_vector(vector):
    """
    Hand-made vector normalization.
    """
    return vector / length_vector(vector)


def length_vector(vector):
    """
    Calculates the norm of a vector.
    """
    return np.linalg.norm(vector)


if __name__ == "__main__":
    pass



from compas_ghpython.artists import NetworkArtist
from compas_ghpython import draw_points

from compas_cem import COLORS


__all__ = ["FormArtist", "TopologyArtist"]


class FormArtist(NetworkArtist):
    """
    """
    def __init__(self, form_diagram, *args, **kwargs):
        super(FormArtist, self).__init__(form_diagram, *args, **kwargs)

        self.edge_colors = {"trail": COLORS["edge"],
                            "deviation": COLORS["edge"]}

        self.edge_state_colors = {-1: COLORS["compression"],
                                  1: COLORS["tension"],
                                  0: COLORS["edge"]}

        self.float_precision = "3f"
        self._form = form_diagram

    @property
    def form(self):
        """
        The form diagram to draw.

        Returns
        -------
        form : :class:`compas_cem.diagrams.FormDiagram`
            A form diagram.
        """
        return self._form

    @form.setter
    def form(self, form):
        self._form = form

    def draw_nodes(self, nodes=None, color=None):
        """Draw a selection of nodes.
        Parameters
        ----------
        nodes: list, optional
            The selection of nodes that should be drawn.
            Default is ``None``, in which case all nodes are drawn.

        Returns
        -------
        list of :class:`Rhino.Geometry.Point3d`
        """
        nodes = nodes or list(self.form.nodes())
        points = []
        for node in nodes:
            points.append({'pos': self.form.node_coordinates(node)})

        return draw_points(points)


class TopologyArtist(NetworkArtist):
    pass



from compas_cem.elements import Edge


class DeviationEdge(Edge):
    """
    A deviation edge.
    """
    def __init__(self, u, v, force):
        super(DeviationEdge, self).__init__(u, v)
        self.attributes = {"force": force, "type": "deviation"}

    def __repr__(self):
        """
        """
        return "{}(force={})".format(self.__class__.__name__, self.attributes["force"])


if __name__ == "__main__":
    pass
